"""Simulates communication with the light controllers via Modbus protocol"""

from pyModbusTCP.server import ModbusServer, DataBank, DataHandler
from pyModbusTCP.client import ModbusClient
from time import sleep
import logging
from prettytable import PrettyTable
import datetime
import argparse
import signal
import data
import sys
import traceback

_RUN_FOR_SEC = 10
_LOG_LEVEL = 20  # INFO

_statistic = {}
_start_time = None
_active = {}


def get_run_options():
    global _RUN_FOR_SEC
    global _LOG_LEVEL
    parser = argparse.ArgumentParser(description='Pass --time')
    parser.add_argument('--time',
                        '-t',
                        type=int,
                        help='Run for the specified interval in seconds',
                        required=False)
    parser.add_argument(
        '--log',
        '-l',
        type=str,
        help='Set the log level like debug, info, warn, error, critical',
        required=False)

    args = parser.parse_args()
    _RUN_FOR_SEC = args.time
    log_level = args.log.upper() if args.log else 'INFO'
    if log_level in logging._nameToLevel:
        _LOG_LEVEL = logging._nameToLevel[log_level]


def run() -> bool:
    global _start_time
    if _RUN_FOR_SEC is None:
        return True
    else:
        return (datetime.datetime.now() -
                _start_time).total_seconds() < _RUN_FOR_SEC


def remove_instance(instance, key: str):
    if type(instance) is ModbusServer:
        instance.stop()
    logging.warning(f'Stopping {key}')
    del _active[key]


def handle_server(instance):
    for reg in data.get_controller_registers(instance.port):
        if not reg['CanOverride']:
            continue
        tp = reg['Type'].lower()
        v = reg['CurrentValue']
        if tp == 'input' and v is not None:
            instance.data_bank.set_input_registers(int(reg['Address']), [v])


def handle_client(instance, ct: dict):
    registers = data.get_controller_registers(instance.port)
    holdingRead = next(
        filter(lambda x: x['Type'] == 'Holding' and x['CanOverride'] is False,
               registers), None)

    holdingWrite = next(
        filter(lambda x: x['Type'] == 'Holding' and x['CanOverride'] is True,
               registers), None)

    regs = {}
    for r in registers:
        regs[r['Address']] = r['Name']
    info = {'Model': ct['Name'], 'Port': instance.port, 'Registers': regs}
    if holdingRead:
        result = instance.read_holding_registers(holdingRead['Address'], 1)
        if not result:
            logging.debug(f'{ct["Name"]} client@{instance.port} no connection')
        else:
            gather('read', 'holding', holdingRead['Address'], instance.port,
                   result[0], info, 'outbound')
            if holdingWrite:
                v = holdingWrite['CurrentValue']
                instance.write_single_register(holdingWrite['Address'], v)
                gather('write', 'holding', holdingWrite['Address'],
                       instance.port, v, info, 'outbound')
    else:
        logging.warning(
            f'{ct["Name"]} client@{instance.port} is not fully configured')


def add_new(controllers: list):
    global _active

    current_active = []
    for ctrl in controllers:
        if ctrl['Enabled']:
            current_active.append(f'{ctrl["Type"]}@{ctrl["Port"]}')
    keys = set(_active.keys())
    added = [x for x in current_active if x not in keys]
    for k in added:
        c = next(
            filter(lambda x: f"{x['Type']}@{x['Port']}" == k, controllers),
            None)
        if not c:
            logging.error(f'Something went bananas for {k}')
            continue
        instance = None
        if c['Model']['RunAs'].lower() == 'server':

            regs = {}
            for r in c['Model']['Registers']:
                regs[r['Address']] = r['Name']
            info = {
                'Model': c['Model']['Name'],
                'Port': c['Port'],
                'Registers': regs
            }
            instance = ModbusServer(c['Ip'],
                                    int(c['Port']),
                                    no_block=True,
                                    data_hdl=MaHandler(data_bank=MaBank(info),
                                                       info=info))
            logging.info(
                f'Starting MODBUS server of type {c["Type"]} at port {instance.port}'
            )
            instance.start()
        else:
            instance = ModbusClient(host=c['Ip'],
                                    port=int(c['Port']),
                                    auto_open=True,
                                    debug=False)
            logging.info(
                f'Starting MODBUS client of type {c["Type"]} at port {instance.port}'
            )
        _active[k] = {
            'instance': instance,
            'type': c['Model'],
        }


def signal_handler(sig, frame):
    global _RUN_FOR_SEC
    logging.warning("KeyboardInterrupt. Shutting down shortly ...")
    _RUN_FOR_SEC = 0


def dispose():
    global _active
    for k, v in _active.items():
        logging.info(f'Shutting down {k} ...')
        if type(v['instance']) is ModbusServer:
            v['instance'].stop()


def gather(op, registry, address, port, value, info, tag='inbound'):
    """Gathers statistics about controllers usage

    Parameters
    ----------
    op : str
        The name (or type) of the operation, like read, write, change, etc
    registry : str
        The type of the register like holding, input, etc
    address : int
        The address of the register
    port : int
        A port (identifier) of incoming connection (not the controller's port)
    value : int or bool
        The value in the register (written or read)
    info : dict
        Contains information about controller's  model, port and registers
    tag : str, optional
        Used to additionally distinguish the operation (default is inbound)

    Returns
    -------
    void
    """

    global _active
    global _statistic
    controller = f'{info["Model"]}@{info["Port"]}'
    try:
        # client port here is a port of incoming connection
        client_key = f'{tag} {port}'
        stat_key = f'{op};{registry} {address};{controller};{value}'

        if client_key in _statistic:
            stats = _statistic[client_key]
            if stat_key in stats:
                stats[stat_key] += 1
            else:
                stats[stat_key] = 1
        else:
            _statistic[client_key] = {stat_key: 1}
    except Exception as e:
        logging.exception(e)

    msg = f'{op} {registry} {controller}:{address} value → {value}'
    if op == 'change event':
        logging.warning(msg)
    elif op == 'write':
        logging.info(msg)
    else:
        logging.debug(msg)


def report():
    global _statistic
    global _start_time

    total_time = (datetime.datetime.now() - _start_time)

    table = PrettyTable()
    table.field_names = [
        'Consumer', 'Operation', 'Registry', 'Controller', 'Value', 'Count'
    ]
    table.align['Consumer'] = "l"
    table.align["Operation"] = "c"
    table.align["Registry"] = "l"
    table.align["Controller"] = "l"
    table.align["Value"] = "c"
    table.align["Count"] = "c"
    table.sortby = "Controller"

    for key, value in _statistic.items():
        for s, c in value.items():
            stat = s.split(';')
            table.add_row([key, stat[0], stat[1], stat[2], stat[3], c])
    print(table)
    print(f'TOTAL RUNTIME {total_time}')


class MaHandler(DataHandler):

    def __init__(self, data_bank, info):
        super().__init__(data_bank=data_bank)
        self.info = info

    def read_coils(self, address, count, srv_info):
        ret = super().read_coils(address, count, srv_info)
        gather('read', 'coil', address, srv_info.client.port, ret.data[0],
               self.info)
        return ret

    def read_d_inputs(self, address, count, srv_info):
        ret = super().read_d_inputs(address, count, srv_info)
        gather('read', 'discrete', address, srv_info.client.port, ret.data[0],
               self.info)
        return ret

    def read_h_regs(self, address, count, srv_info):
        ret = super().read_h_regs(address, count, srv_info)
        gather('read', 'holding', address, srv_info.client.port, ret.data[0],
               self.info)
        return ret

    def read_i_regs(self, address, count, srv_info):
        ret = super().read_i_regs(address, count, srv_info)
        gather('read', 'input', address, srv_info.client.port, ret.data[0],
               self.info)
        return ret

    def write_coils(self, address, bits_l, srv_info):
        ret = super().write_coils(address, bits_l, srv_info)
        gather('write', 'coils', address, srv_info.client.port, bits_l[0],
               self.info)
        self.update_reg(address, bits_l[0])
        return ret

    def write_h_regs(self, address, words_l, srv_info):
        ret = super().write_h_regs(address, words_l, srv_info)
        gather('write', 'holding', address, srv_info.client.port, words_l[0],
               self.info)
        self.update_reg(address, words_l[0])
        return ret

    def update_reg(self, address, value):
        data.set_controller_register(self.info['Model'], self.info['Port'],
                                     self.info['Registers'][address], value)


class MaBank(DataBank):

    def __init__(self, info):
        super().__init__()
        self.info = info

    def on_coils_change(self, address, from_value, to_value, srv_info):
        gather('change event', 'coils', address, srv_info.client.port,
               f'{from_value}⟶{to_value}', self.info)

    def on_holding_registers_change(self, address, from_value, to_value,
                                    srv_info):
        gather('change event', 'holding', address, srv_info.client.port,
               f'{from_value}⟶{to_value}', self.info)


if __name__ == '__main__':
    try:
        signal.signal(signal.SIGINT, signal_handler)
        get_run_options()
        logging.basicConfig(format='%(levelname)s %(message)s',
                            level=_LOG_LEVEL)
        _start_time = datetime.datetime.now()

        version = data.get_version()
        logging.info(f'abibas v{version}. Light controllers MODBUS simulator')

        while run():
            controllers = data.get_controllers()

            add_new(controllers)

            for k in list(_active.keys()):
                instance = _active[k]['instance']
                c = next(
                    filter(lambda c: instance.port == int(c['Port']),
                           controllers), None)

                if not c or not c['Enabled']:
                    remove_instance(instance, k)
                    continue

                if type(instance) is ModbusServer:
                    handle_server(instance)
                else:
                    handle_client(instance, c['Model'])
            sleep(5)

        report()
        dispose()

    except Exception:
        exc = sys.exception()
        traceback.print_tb(exc.__traceback__, limit=3, file=sys.stdout)
        logging.critical('An epic failure just has happened!')
        dispose()
